/*
 * main.c
 * Thomas Nemeth -- le 13.03.2005
 *
 *   Copyright (C) 1999  Thomas Nemeth
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <ctype.h>
#include <cdaudio.h>
#include "config.h"
#include "xmem.h"
#include "main.h"


#define DEFAULT_DEVICE "/dev/cdrom"

#define USE_FREEDB  1
#define USE_CDINDEX 2
/*
#ifndef USE_FREEDB
#  define USE_CDINDEX
#endif
*/


static char *device = NULL;
static char *output = NULL; /* unused */
static char *request_track = NULL;
static int track = 0;
static int start_pos = 0;
static int cd_fd = -1;
static int verbose = 0;
static int total;
static int list = 0;
static int force = 0;
static int cddb_ok = 1;
static int play_all = 0;
static int cddb_method = USE_FREEDB;
static struct disc_info infos;
static struct disc_data datas;



static void signals_handler(int sig)
{
        cd_stop(cd_fd);
        cd_finish(cd_fd);
        if (verbose)
        {
                printf("\n");
        }
        exit(sig + 10);
}


static int get_track(const char *option)
{
        int ret = 0;

        ret = atoi(option);
        if (ret == 0)
        {
                char *data = xstrdup(option);
                int i = 0;
                while (data[i] && !isdigit(data[i]))
                {
                        i++;
                }
                if (data[i])
                {
                        ret = atoi(&data[i]);
                }
                free(data);
        }
        if (ret == 0)
        {
                request_track = xstrdup(option);
                ret = -1;
        }

        return ret;
}


static void usage(int version)
{
        if (version)
        {
                printf(PROGRAM" v"VERSION" Thomas Nemeth, 2005\n");
        }
        else
        {
                printf("Usage : cda123 [-h] [-V] [-v] [-l] [-f]"
                       " [-d device] [-k pos] [-a] [-c] track\n"
                       " -h : this help\n"
                       " -V : show version and exit\n"
                       " -v : verbose mode\n"
                       " -l : list tracks and exit\n"
                       " -f : force play even if drive does not report status\n"
                       " -d : set device (defaults : /dev/cdrom)\n"
                       " -k : set the starting position (seconds)\n"
                       " -a : play all tracks from specified track\n"
                       " -c : use cdindex instead of freedb\n"
                       " track : track to play (starting from 1)\n\n");
                printf("This program is created for use in TMMS. That's why it"
                       " displays tracks names\nwith .cda at the end. Anyway"
                       " you can still use it as a standalone player.\n");
        }
        exit(0);
}


static void parse_args(int argc, const char *argv[])
{
#define ERROR "%s : error, unknown option \"%s\"\n", argv[0], argv[option]
        int option = 1;

        while (option < argc)
        {
                if (argv[option][0] == '-')
                {
                        switch (argv[option][1])
                        {
                                case 'h':
                                        usage(FALSE);
                                        break;
                                case 'V':
                                        usage(TRUE);
                                        break;
                                case 'o':
                                        output = xstrdup(argv[++option]);
                                        break;
                                case 'd':
                                        device = xstrdup(argv[++option]);
                                        break;
                                case 'v':
                                        verbose = 1;
                                        break;
                                case 'k':
                                        start_pos = atoi(argv[++option]);
                                        break;
                                case 'l':
                                        list = 1;
                                        break;
                                case 'f':
                                        force = 1;
                                        break;
                                case 'a':
                                        play_all = 1;
                                        break;
                                case 'c':
                                        cddb_method = USE_CDINDEX;
                                        break;
                                default:
                                        fprintf(stderr, ERROR);
                        }
                }
                else
                {
                        if (track == 0)
                        {
                                track = get_track(argv[option]);
                        }
                        else
                        {
                                fprintf(stderr, ERROR);
                        }
                }
                option++;
        }
#undef ERROR
}


static void display_time()
{
        int elapsed = infos.disc_track_time.minutes * 60
                + infos.disc_track_time.seconds;
        int remaining = total - elapsed;
        int rmins = remaining / 60;
        int rsecs = remaining - (rmins * 60);

        printf("\rTime: %02d:%02d.%02d [%02d:%02d]   ",
                infos.disc_track_time.minutes,
                infos.disc_track_time.seconds,
                infos.disc_track_time.frames,
                rmins,
                rsecs);
        fflush(stdout);
}


static void track_infos()
{
        int mins = infos.disc_track[track - 1].track_length.minutes;
        int secs = infos.disc_track[track - 1].track_length.seconds;

        if (verbose)
        {
                char *track_name = NULL;

                if (cddb_ok == 0)
                {
                        track_name = xmalloc(10);
                        snprintf(track_name, 10, "Track %d", track);
                }
                else
                {
                        char *name = datas.data_track[track - 1].track_name;
                        if ((name[0] != 0) && (strcmp(name, "Unknown") != 0))
                        {
                                track_name = xstrdup(name);
                        }
                        else
                        {
                                track_name = xmalloc(10);
                                snprintf(track_name, 10, "Track %d", track);
                        }
                }
                printf("[%02d:%02d] %s.cda\n", mins, secs, track_name);
                free(track_name);
                if (!list)
                {
                        printf("Playing track %d\n", track);
                        printf("Starting at position %d\n", start_pos);
                }
        }
        total = mins * 60 + secs;
}


static void list_cd_tracks()
{
        verbose = 1;
        if (cddb_ok == 1)
        {
                printf(" Artist : %s\n"
                       " Album  : %s\n"
                       " Genre  : %s\n",
                       datas.data_artist,
                       datas.data_title,
                       cddb_genre(datas.data_genre));
        }
        printf(" Total tracks : %d\n", infos.disc_total_tracks);
        printf(" Disc length  : %02d:%02d\n",
                        infos.disc_length.minutes,
                        infos.disc_length.seconds);
        for (track = 1 ; track <= infos.disc_total_tracks ; track++)
        {
                track_infos();
        }
        exit(0);
}


static void lookup_track()
{
        for (track = 1 ; track <= infos.disc_total_tracks ; track++)
        {
                char *track_name = datas.data_track[track - 1].track_name;
                if ((strstr(track_name, request_track) != NULL) ||
                    (strstr(request_track, track_name) != NULL))
                {
                        break;
                }
        }
        free(request_track);
}


static void init_device()
{
        if (device == NULL)
        {
                device = xstrdup(DEFAULT_DEVICE);
        }
        cd_fd = cd_init_device(device);
        signal(SIGINT, signals_handler);
        signal(SIGQUIT, signals_handler);
        signal(SIGTERM, signals_handler);
        cd_stat(cd_fd, &infos);
        if (infos.disc_present == 0)
        {
                fprintf(stderr, "%s : No disc in drive...\n", device);
                cd_finish(cd_fd);
                exit(2);
        }
}


static void init_cddb_infos()
{
        struct cddb_conf       cddb_conf;
        struct cddb_serverlist server_list;
        struct cddb_host       proxy_host;

        cddb_conf.conf_access = CDDB_ACCESS_REMOTE;
        if (cddb_method == USE_FREEDB)
        {
                server_list.list_len = 1;
                strncpy(server_list.list_host[0].host_server.server_name,
                                "www.freedb.org", 256);
                strncpy(server_list.list_host[0].host_addressing,
                                "cgi-bin/cddb.cgi", 256);
                server_list.list_host[0].host_server.server_port = 80;
                server_list.list_host[0].host_protocol = CDDB_MODE_HTTP;
        }
        else
        {
                server_list.list_len = 1;
                strncpy(server_list.list_host[0].host_server.server_name,
                                "www.cdindex.org", 256);
                strncpy(server_list.list_host[0].host_addressing,
                                "cgi-bin/cdi/get.pl", 256);
                server_list.list_host[0].host_server.server_port = 80;
                server_list.list_host[0].host_protocol = CDINDEX_MODE_HTTP;
        }
        cddb_write_serverlist(
                        cddb_conf,
                        server_list,
                        proxy_host.host_server);

        if (cddb_read_disc_data(cd_fd, &datas) < 0)
        {
                cddb_ok = 0;
        }
}


static void check_track()
{
        if (track == -1)
        {
                lookup_track();
        }
        track += infos.disc_first_track - 1;
        if (track > infos.disc_total_tracks)
        {
                fprintf(stderr, "Track number is too big : max = %d\n",
                                infos.disc_total_tracks);
                cd_finish(cd_fd);
                exit(3);
        }
        if (infos.disc_track[track - 1].track_type != CDAUDIO_TRACK_AUDIO)
        {
                fprintf(stderr, "Track is data track."
                                " Please select an audio track.\n");
                
                cd_finish(cd_fd);
                exit(4);
        }
}


static void cda123_init()
{
        printf("cda123 version "VERSION" ready...\n");
        if ((track == 0) && (list == 0))
        {
                fprintf(stderr, "You have to specify a track to play !\n");
                exit(1);
        }
        init_device();
        init_cddb_infos();
        if (list == 1)
        {
                list_cd_tracks();
        }
        check_track();
        track_infos();
}


static void play_track()
{
        int err;

        if (play_all)
	{
	        err = cd_play_pos(cd_fd, track, start_pos);
	}
        else
        { 
                err = cd_play_track_pos(cd_fd, track, track, start_pos);
        }
        if (err == 0)
        {
                cd_stat(cd_fd, &infos);
                if (force)
                {
                        goto out;
                }
                while ((infos.disc_mode != CDAUDIO_COMPLETED)
                        && (infos.disc_mode != CDAUDIO_NOSTATUS))
                {
                        /* 50 ms should be enough */
                        struct timespec ts = {0, 50000};

                        nanosleep(&ts, NULL);
                        cd_stat(cd_fd, &infos);
                        if (verbose)
                        {
                                display_time();
                        }
                }
                if (verbose)
                {
                        printf("\n");
                }
                if (infos.disc_mode == CDAUDIO_NOSTATUS)
                {
                        fprintf(stderr, "Error : unable to read CD status\n");
                }
        }
        else
        {
                fprintf(stderr, "Error : %d\n", err);
        }
out:
}


int main(int argc, const char *argv[])
{
        parse_args(argc, argv);
        cda123_init();
        play_track();
        cd_stop(cd_fd);
        cd_finish(cd_fd);
        FREE(device);
        FREE(output);

        return 0;
}

